from typing import Any

import aiomqtt

from . import MqttClient, MqttConfig, MqttError, SpbMessage, Topic, Will
from .encoding import JsonEncoder, ProtobufEncoder


class PahoMqttClient(MqttClient):
    def __init__(
        self,
        config: dict[str, Any],
    ):
        config["port"] = int(config["port"])
        config["keepalive"] = int(config.get("keepalive", 30))
        self._config = MqttConfig(**config)

        self._client = None
        self._json_encoder = JsonEncoder()
        self._protobuf_encoder = ProtobufEncoder()
        self._is_host = None

    async def connect(self, component_name: str, will: Will):
        payload = self._encode_message(will.message)
        will_ = aiomqtt.Will(
            will.message.topic.value, payload, qos=will.qos, retain=will.retain
        )

        tls_pars = None
        if self._config.ca_certs:
            tls_pars = aiomqtt.TLSParameters(ca_certs=self._config.ca_certs)

        self._client = aiomqtt.Client(
            self._config.hostname,
            self._config.port,
            identifier=component_name,
            username=self._config.username,
            password=self._config.password,
            will=will_,
            protocol=aiomqtt.ProtocolVersion.V5,
            tls_params=tls_pars,
            clean_start=True,
            keepalive=self._config.keepalive,
        )

        try:
            await self._client.__aenter__()  # Connect with aiomqtt
        except Exception as e:
            self._client = None
            raise MqttError() from e

    @property
    def keepalive(self) -> float:
        return self._config.keepalive

    @property
    def is_connected(self) -> bool:
        return self._client is not None

    def _encode_message(self, message: SpbMessage) -> bytes:
        if message.topic.value.startswith("spBv1.0/STATE/"):
            return self._json_encoder.encode(message.payload)

        return self._protobuf_encoder.encode(message.payload)

    async def publish(self, message: SpbMessage, qos: int, retain: bool):
        if self._client is None:
            raise RuntimeError("Client not connected to Mqtt Server")

        try:
            await self._client.publish(
                message.topic.value,
                self._encode_message(message),
                qos=qos,
                retain=retain,
            )
        except Exception as e:
            self._client = None
            raise MqttError() from e

    async def subscribe(self, topic: str, qos: int):
        if self._client is None:
            raise RuntimeError("Client not connected to Mqtt Server")

        try:
            await self._client.subscribe(topic, qos=qos)
        except Exception as e:
            self._client = None
            raise MqttError() from e

    async def deliver_message(self) -> SpbMessage:
        if self._client is None:
            raise RuntimeError("Mqtt client is not connected")

        try:
            message = await anext(self._client.messages)
        except Exception as e:
            self._client = None
            raise MqttError() from e

        # if type(message.payload) is not bytes:
        #     raise ValueError(f"Message should be bytes {message.payload}")

        if message.topic.value.startswith("spBv1.0/STATE/"):
            return SpbMessage(
                Topic(message.topic.value), self._json_encoder.decode(message.payload)
            )

        message = SpbMessage(
            Topic(message.topic.value), self._protobuf_encoder.decode(message.payload)
        )

        return message

    async def disconnect(self):
        if self._client is None:
            return

        await self._client.__aexit__(None, None, None)
        self._client = None



"""Edge Nodes are implemented following the State Design Pattern.
All states are inside this module
"""
import abc
import asyncio
import logging
from asyncio import create_subprocess_shell
from dataclasses import dataclass

from aiospb.nodes.messages import MqttNodeCarrier, NodePayload

from ..data import DataType, Metric
from ..mqtt import MqttError
from .scanning import Scanner
from .stores import HistoricalStore

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class NodeState(abc.ABC):
    """Interface to be followed by any node state"""

    @abc.abstractmethod
    async def establish_session(self):
        """Called by edge node"""

    @abc.abstractmethod
    async def terminate_session(self):
        """Called by edge node"""


@dataclass
class NodeContext:
    """Group all common state variables into one"""

    carrier: MqttNodeCarrier
    scanner: Scanner
    store: HistoricalStore
    primary_hostname: str = ""
    state: NodeState | None = None


class Disconnected(NodeState):
    """Edge is disconnected to MQTT"""

    def __init__(self, context: NodeContext):
        self._context = context
        context.state = self
        self._waiting_to_host_reporting = None
        self._saving_changes = asyncio.create_task(self._save_changes())
        logger.info("Edge Node is DISCONNECTED")

    async def establish_session(self):
        trys = 0
        while True:
            try:
                await self._context.carrier.connect()
                self._waiting_to_host_reporting = asyncio.create_task(
                    self._when_host_is_reporting()
                )
                return
            except MqttError as e:
                trys += 1
                if trys < 7:
                    wait = 10
                elif trys < 12:
                    wait = 60
                else:
                    wait = 300
                logger.warning(f"Not possible to connect, waiting {wait} seconds")
                await asyncio.sleep(wait)

    async def _when_host_is_reporting(self):
        while True:
            try:
                message = await self._context.carrier.deliver_message()
            except MqttError as e:
                logger.error("When recieving messages from host, restablishing session")
                logger.exception(e)
                await self.establish_session()
                return

            if not message.is_a("STATE"):
                continue

            if (
                self._context.primary_hostname
                and message.topic.value
                != f"spBv1.0/STATE/{self._context.primary_hostname}"
            ):
                continue

            if message.payload.to_dict()["online"]:
                self._saving_changes.cancel()
                birth = await self._context.scanner.start()
                logger.debug(f"Found {len(birth)} root metrics in the Edge Node")
                try:
                    await self._context.carrier.send_birth(birth)
                    logger.info("Birth certificate published to MQTT server")
                    self._waiting_to_host_reporting = None
                    Reporting(self._context)
                except MqttError as e:
                    logger.error(
                        "When sending birth certificate, restablishing session"
                    )
                    logger.exception(e)
                    await self.establish_session()
                finally:
                    return

    async def _save_changes(self):
        while True:
            changes = await self._context.scanner.deliver_changes()
            logger.info(f"Saving {len(changes)} changes as history...")

            try:
                await self._context.store.save_dtos(changes)
                logger.info("Saved changes to temporary store")
            except Exception as e:
                logger.warning("Not saved the samples in history!!!")
                logger.exception(e)

    async def terminate_session(self):
        """Close session, stopping scanning"""
        self._cancel_tasks()

        await self._context.scanner.stop()
        if self._context.carrier.is_connected:
            logger.debug("Disconnecting cleanly ...")
            await self._context.carrier.send_death()
            await self._context.carrier.disconnect()
            logger.info("Edge node cleanly disconnected")

    def _cancel_tasks(self):
        if self._waiting_to_host_reporting is not None:
            self._waiting_to_host_reporting.cancel()
        self._saving_changes.cancel()


class Reporting(NodeState):
    """Gateway is connected to MQTT Server and Primary Host is Reporting"""

    def __init__(self, context: NodeContext):
        self._context = context
        context.state = self
        self._sending = asyncio.create_task(self._send_node_changes())
        self._recieving = asyncio.create_task(self._recieve_host_messages())
        logger.info("Edge Node is REPORTING")

    async def establish_session(self):
        """Not allowed, only when is disconnected to Mqtt Broker"""
        raise RuntimeError(
            f'State "REPORTING" can not establish session, terminate_session before'
        )

    async def terminate_session(self):
        """Terminate cleanly the session"""
        logger.info(f"Terminating session REPORTING")
        self._cancel_tasks()
        await Disconnected(self._context).terminate_session()

    def _cancel_tasks(self):
        """Stops tasks, cancel incomming, wait messages to be sent."""
        self._sending.cancel()
        self._recieving.cancel()

    async def _send_node_changes(self):
        while True:
            changes = await self._context.scanner.deliver_changes()
            logger.info(f"Detected {len(changes)} changes from the scanner")

            if changes:
                history = []
                try:
                    if self._context.store.has_dtos():
                        history = await self._context.store.load_all_dtos()
                        logger.info(f"Loaded {len(history)} historical changes")
                except Exception as e:
                    logger.error("Not possible to load history of changes")
                    logger.exception(e)

                try:
                    await self._context.carrier.send_data_changes(history + changes)
                    logger.info("Sent changes to the MQTT Server")
                except MqttError as e:
                    logger.error("MQTT error when sending changes")
                    try:
                        await self._context.store.save_dtos(changes)
                    except Exception as e:
                        logger.error("Not possible to save changes")
                        logger.exception(e)
                    self._cancel_tasks()
                    await Disconnected(self._context).establish_session()
                    return

            try:
                if self._context.store.has_dtos():
                    await self._context.store.clear()
            except Exception as e:
                logger.error("Not possible to clean scan changes from store")
                logger.exception(e)

    async def _recieve_host_messages(self):
        while True:
            try:
                message = await self._context.carrier.deliver_message()
            except MqttError as e:
                logger.error(f"Error when recieving host messages")
                logger.exception(e)
                self._cancel_tasks()
                await Disconnected(self._context).establish_session()
                return

            if message.is_a("NCMD"):
                if type(message.payload) is NodePayload:
                    await self._process_commands(message.payload)
                continue

            if (
                self._context.primary_hostname
                and message.topic.value
                != f"spBv1.0/STATE/{self._context.primary_hostname}"
            ):
                continue

            if not message.payload.to_dict()["online"]:
                logger.info(f"Detected host is disconnected, stopped data sending")
                self._cancel_tasks()
                Connected(self._context)
                return

    async def _process_commands(self, payload: NodePayload):
        commands = payload.metrics
        confirmations = []
        for index, command in enumerate(commands):
            if command.name == "Node Control/Reboot":
                confirmations.append(
                    Metric(0, True, DataType.Boolean, name="Node Control/Reboot")
                )
                if index + 1 < len(commands):
                    logger.warning(
                        f"The commands {commands[index+1:]} will not applied after reboot"
                    )
                await self._context.carrier.confirm_commands(confirmations)
                await self._reboot()
                return

            elif command.name == "Node Control/Rebirth":
                logger.debug("Recieved rebirth command from host")
                try:
                    await self._rebirth()
                except MqttError as e:
                    logger.error("MQTT connection error when rebirth")
                    logger.exception(e)
                    self._cancel_tasks()
                    await Disconnected(self._context).establish_session()

                confirmations.append(
                    Metric(0, True, DataType.Boolean, name="Node Control/Rebirth")
                )
            else:
                description = command.name + ":" + str(command.alias)
                logger.info(f"Writing to metric '{description}'")

                confirmations.append(
                    await self._context.scanner.execute_command(command)
                )

        if confirmations:
            await self._context.carrier.confirm_commands(confirmations)
            logger.info(f"Sent confirmations of {len(confirmations)} changes")

    async def _reboot(self):
        await self.terminate_session()
        process = await create_subprocess_shell(
            "reboot", stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
        )

        # Wait for the command to complete
        await process.communicate()

    async def _rebirth(self):
        await self._context.scanner.stop()
        birth = await self._context.scanner.start()
        logger.debug(f"Found {len(birth)} root metrics for birth certificate")
        await self._context.carrier.send_birth(birth)
        logger.info(f"Resent birth certificate to MQTT server")


class Connected(NodeState):
    """Gateway connected to MQTT Server, but primary host is connected"""

    def __init__(self, context: NodeContext):
        self._context = context
        context.state = self
        self._recieving_messages = asyncio.create_task(self._recieve_host_messages())
        self._saving_changes = asyncio.create_task(self._save_changes())
        logger.info("Edge Node is CONNECTED")

    async def establish_session(self):
        """Not allowed, only when is disconnected to Mqtt Broker"""

        raise RuntimeError(
            f'State "connected" can not establish session, terminate_session before'
        )

    async def terminate_session(self):
        """Close session of node cleanly"""
        logger.info(f"Terminating session from CONNECTED state")
        self._cancel_tasks()
        await Disconnected(self._context).terminate_session()

    def _cancel_tasks(self):
        self._recieving_messages.cancel()
        self._saving_changes.cancel()

    async def _save_changes(self):
        while True:
            changes = await self._context.scanner.deliver_changes()
            logger.info(f"Saving {len(changes)} changes as history...")

            try:
                await self._context.store.save_dtos(changes)
                logger.info("Saved changes to temporary store")
            except Exception as e:
                logger.warning("Not saved the samples in history!!!")
                logger.exception(e)

    async def _recieve_host_messages(self):
        while True:
            try:
                message = await self._context.carrier.deliver_message()
            except MqttError as e:
                logger.error("Lost MQTT connection when recieving host messages")
                logger.exception(e)
                self._cancel_tasks()
                await Disconnected(self._context).establish_session()
                return

            if not message.is_a("STATE"):
                continue

            if (
                self._context.primary_hostname
                and message.topic.value
                != f"spBv1.0/STATE/{self._context.primary_hostname}"
            ):
                continue

            if message.payload.to_dict()["online"]:
                logger.info("The host application is ONLINE!!")
                await self._context.scanner.stop()
                birth = await self._context.scanner.start()
                try:
                    await self._context.carrier.send_birth(birth)
                    logger.info("Birth certificate is sent")
                except MqttError:
                    await Disconnected(self._context).establish_session()

                self._cancel_tasks()
                Reporting(self._context)
                return



import abc
import asyncio
import logging
from dataclasses import dataclass, field
from typing import Any, Self

from .. import Clock, UtcClock
from ..data import DataType, Metric, PropertySet, PropertyValue, Quality, ValueType

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class MetricCore:
    """Metric data, stable for a session"""

    name: str
    datatype: DataType
    properties: PropertySet = field(default_factory=PropertySet)
    alias: int = 0
    is_transient: bool = False

    def create_metric(
        self, value: ValueType, timestamp: int, quality: Quality | None = None
    ) -> Metric:
        """Create metric from the metric core"""
        props = self.properties
        if quality is not None:
            props_data = self.properties.as_dict()
            props_data["quality"] = {"value": quality.value, "dataType": "Int32"}
            props = PropertySet.from_dict(props_data)

        return Metric(
            timestamp,
            value,
            self.datatype,
            self.alias,
            self.name,
            props,
            self.is_transient,
        )

    # @classmethod
    # def from_dict(cls, dump: dict[str, Any]) -> Self:
    #     datatype = (
    #         DataType(dump["dataType"])
    #         if type(dump["dataType"]) is int
    #         else DataType[dump["dataType"]]
    #     )
    #     properties = (
    #         PropertySet.from_dict(dump["properties"])
    #         if "properties" in dump
    #         else PropertySet()
    #     )
    #     return cls(
    #         dump["name"],
    #         datatype,
    #         properties,
    #         dump.get("alias", 0),
    #         dump.get("is_transient", False),
    #     )


class MetricsNet(abc.ABC):
    # @abc.abstractmethod
    # async def smap(self) -> dict[str, int]:
    #     """List all name of metrics available and their aliases, 0 it does not have"""

    @abc.abstractmethod
    async def get_metric_cores(self, filter: str = "") -> list[MetricCore]:
        """List available metrics in the net with all its properties"""

    @abc.abstractmethod
    async def read_value(self, metric_name: str = "", alias: int = 0) -> ValueType:
        """Read a value to a metric"""

    @abc.abstractmethod
    async def write_value(
        self, value: ValueType, metric_name: str = "", alias: int = 0
    ):
        """Write a value to a metric"""


@dataclass
class _Sample:
    timestamp: int
    value: ValueType
    quality: Quality

    @classmethod
    async def read(
        cls,
        metrics_net: MetricsNet,
        clock: Clock,
        metric_name: str = "",
        alias: int = 0,
    ) -> Self:
        if not alias and not metric_name:
            raise ValueError("Metric Name and alias can not be null at the same time")
        try:
            value = await metrics_net.read_value(metric_name)
            return cls(clock.timestamp(), value, Quality.GOOD)
        except asyncio.CancelledError:
            return cls(clock.timestamp(), None, Quality.STALE)
        except Exception as e:
            logger.error(f"Exception when reading metric {metric_name}:{alias}")
            logger.exception(e)
            return cls(clock.timestamp(), None, Quality.BAD)


@dataclass
class _Scan:
    scan_rate: int
    datatype: DataType
    quality: Quality
    value: ValueType = None
    alias: int = 0
    name: str = ""

    def update(self, sample: _Sample) -> Metric | None:
        if sample.quality == self.quality and sample.value == self.value:
            return

        if self.quality != Quality.GOOD and sample.quality == self.quality:
            return

        props = (
            PropertySet()
            if self.quality == sample.quality
            else PropertySet(
                ("quality",), (PropertyValue(sample.quality.value, DataType.Int32),)
            )
        )
        self.quality = sample.quality

        value = sample.value if sample.quality == Quality.GOOD else self.value
        self.value = value

        name = "" if self.alias else self.name
        return Metric(sample.timestamp, value, self.datatype, self.alias, name, props)

    def plan_next_scan(self, schedules: dict[int, list[Self]], clock: Clock):
        timestamp = ((clock.timestamp() // self.scan_rate) + 1) * self.scan_rate
        if timestamp not in schedules:
            schedules[timestamp] = []

        schedules[timestamp].append(self)


class Scanner:
    def __init__(
        self,
        metrics_net: MetricsNet,
        scan_rate: int = 60000,
        clock: Clock | None = None,
    ):
        self._net = metrics_net
        self._clock = clock or UtcClock()
        self._scan_rate = scan_rate

        self._queue = asyncio.Queue()
        self._schedules = {}  # scans ordered by timestamp to be triggered
        self._tasks = {}  # scannings to be done in the future

    @property
    def scan_rate(self) -> float:
        return self._scan_rate

    async def deliver_changes(self, timeout: float = 0.0) -> list[Metric]:
        """Return last changes"""
        if timeout:
            return await asyncio.wait_for(self._queue.get(), timeout=timeout)
        return await self._queue.get()

    async def _read_births(self) -> list[Metric]:
        cores = await self._net.get_metric_cores()

        readings = [
            asyncio.create_task(_Sample.read(self._net, self._clock, core.name))
            for core in cores
        ]
        _, pending = await asyncio.wait(readings, timeout=self._scan_rate / 1000)
        for reading in pending:
            reading.cancel()
        await asyncio.gather(*pending, return_exceptions=True)

        births = []
        for core, reading in zip(cores, readings):
            sample = reading.result()
            births.append(
                core.create_metric(sample.value, sample.timestamp, sample.quality)
            )

        return births

    async def start(self) -> list[Metric]:
        ts = self._clock.timestamp()
        node_controls = [
            Metric(ts, False, DataType.Boolean, name="Node Control/Reboot"),
            Metric(ts, False, DataType.Boolean, name="Node Control/Rebirth"),
            Metric(
                ts,
                int(self._scan_rate * 1000),
                DataType.Int64,
                name="Node Control/Scan Rate",
            ),
        ]

        births = await self._read_births()
        scans = []
        for birth in births:
            if "Properties/" in birth.name:  # Properties will not be scanned
                continue

            scan_rate = self._scan_rate
            if "scan_rate" in birth.properties:
                scan_rate = birth.properties["scan_rate"].value
                if type(scan_rate) is not int:
                    raise ValueError(f"Scan rate can not be of type {type(scan_rate)}")

                if scan_rate == 0:  # No scan this metric
                    continue

            scans.append(
                _Scan(
                    scan_rate,
                    birth.datatype,
                    Quality(birth.properties["quality"].value),
                    birth.value,
                    alias=birth.alias,
                    name=birth.name,
                )
            )

        self._plan_scans(scans)

        return node_controls + births

    async def _scan_schedule(self, timestamp: int):
        wait_time = (timestamp - self._clock.timestamp()) / 1000
        await self._clock.asleep(wait_time if wait_time > 0 else 0)

        readings = await self._read_metrics(self._schedules[timestamp])
        self._report_changes(readings)
        self._plan_scans(self._schedules.pop(timestamp, []))
        if timestamp in self._tasks:
            self._tasks.pop(timestamp)

    async def _read_metrics(self, scans: list[_Scan]) -> list:
        readings = [
            asyncio.create_task(
                _Sample.read(self._net, self._clock, scan.name, scan.alias)
            )
            for scan in scans
        ]
        logger.debug(f"Scanning {len(readings)} metrics")

        timeout = self._scan_rate / 2000  #  half of scan rate
        try:
            _, pending = await asyncio.wait(
                readings,
                timeout=timeout,
            )
        except asyncio.CancelledError as e:
            for reading in readings:
                if not reading.done():
                    reading.cancel()
            raise e

        if pending:
            logger.warning(f"There are {len(pending)} staled readings by timeout")

        for reading in pending:
            reading.cancel()
        await asyncio.gather(*pending)

        changes = []
        for scan, reading in zip(scans, readings):
            metric = scan.update(reading.result())
            if metric is not None:
                changes.append(metric)

        return sorted(changes, key=lambda change: change.timestamp)

    def _report_changes(self, changes):
        if changes:
            self._queue.put_nowait(changes)
            logger.info(f"Reported {len(changes)} changes after scanning")
        else:
            logger.info("No changes to be reported after scanning")

    def _plan_scans(self, scans):
        for scan in scans:
            scan.plan_next_scan(self._schedules, self._clock)

        for timestamp in self._schedules.keys():
            if timestamp not in self._tasks:
                self._tasks[timestamp] = asyncio.create_task(
                    self._scan_schedule(timestamp)
                )

        logger.info(f"Planned to scan {len(scans)} metrics")

    async def stop(self):
        """Stop all next scanning tasks"""
        for task in self._tasks.values():
            task.cancel()
        await asyncio.gather(*list(self._tasks.values()), return_exceptions=True)
        self._tasks.clear()
        self._schedules.clear()

    def _find_scan(self, metric: Metric) -> _Scan | None:
        for scans in self._schedules.values():
            for scan in scans:
                if (metric.alias and metric.alias == scan.alias) or (
                    metric.name and metric.name == scan.metric_name
                ):
                    return scan

    async def execute_command(self, metric: Metric) -> Metric:
        """execute a command, writing metric to the net"""

        if metric.name == "Node Control/Scan Rate":
            if type(metric.value) is not int:
                raise ValueError(f"Metric {metric.name} has not type integer")
            self.scan_rate = metric.value
            return Metric(
                self._clock.timestamp(),
                metric.value,
                DataType.Int32,
                name="Node Control/Scan Rate",
            )

        try:
            await self._net.write_value(
                metric.value, metric_name=metric.name, alias=metric.alias
            )
            sample = _Sample(self._clock.timestamp(), metric.value, Quality.GOOD)
        except asyncio.CancelledError:
            sample = _Sample(self._clock.timestamp(), metric.value, Quality.STALE)
        except Exception as e:
            logger.error(
                f"Writing {metric.value} to metric '{metric.name}:{metric.alias}'"
            )
            logger.exception(e)
            sample = _Sample(self._clock.timestamp(), metric.value, Quality.BAD)

        scan = self._find_scan(metric)
        if scan:
            scan.update(sample)
        return Metric(
            sample.timestamp,
            metric.value,
            metric.datatype,
            metric.alias,
            metric.name,
            properties=PropertySet.from_dict(
                {"quality": {"value": sample.quality.value, "dataType": "Int32"}}
            ),
        )

    @scan_rate.setter
    def scan_rate(self, value: int):
        replanned_scans = []
        for timestamp in list(self._schedules.keys()):
            scans = []
            for scan in self._schedules.pop(timestamp):
                if scan.scan_rate == self._scan_rate:
                    replanned_scans.append(
                        _Scan(
                            value,
                            scan.datatype,
                            scan.quality,
                            scan.value,
                            scan.alias,
                            scan.name,
                        )
                    )
                else:
                    scans.append(scan)
            if scans:
                self._schedules[timestamp] = scans
            else:
                self._tasks.pop(timestamp).cancel()

        self._plan_scans(replanned_scans)
        self._scan_rate = value



__all__ = (
    "EdgeNode",
    "OnmemHistoricalStore",
    "FsHistoricalStore",
    "MetricsNet",
    "MetricCore",
    "Metric",
    "ValueType",
)

from ..data import Metric, ValueType
from .nodes import EdgeNode
from .scanning import MetricCore, MetricsNet
from .stores import FsHistoricalStore, OnmemHistoricalStore



from aiospb import Clock, UtcClock
from aiospb.nodes.ndata import Metric
from aiospb.nodes import MetricsNet


class NodeCommander:
    def __init__(self, metrics_net: MetricsNet, clock: Clock | None):
        self._net = metrics_net
        self._clock = clock or UtcClock()

    async def write_metric(self, metric: Metric):
        """Write a metric to the net"""



import abc
import csv
import os.path

import aiofiles
from aiofiles import open

from aiospb.data import Metric

from ..data import DataType, PropertySet, PropertyValue


class HistoricalStore(abc.ABC):
    """Store historical changes to retrieve when host application is online"""

    @abc.abstractmethod
    async def save_dtos(self, changes: list[Metric]):
        """Save all change metrics of a scan"""

    @abc.abstractmethod
    async def load_all_dtos(self) -> list[Metric]:
        """Load all historical metric changes"""

    @abc.abstractmethod
    def has_dtos(self) -> bool:
        """Has data stored?"""

    @abc.abstractmethod
    async def clear(self):
        """Clear content of historical"""


class OnmemHistoricalStore(HistoricalStore):
    """Store historical changes on memory. Not recomemded if host will be stopped for a long time"""

    def __init__(self):
        self._dtos = []

    async def save_dtos(self, dtos: list[Metric]):
        self._dtos.extend(dtos)

    async def load_all_dtos(self) -> list[Metric]:
        dtos = [
            Metric(
                dto.timestamp,
                dto.value,
                dto.datatype,
                dto.alias,
                dto.name,
                dto.properties,
                is_historical=True,
            )
            for dto in self._dtos
        ]
        self._dtos.clear()
        return dtos

    def has_dtos(self):
        return bool(self._dtos)

    async def clear(self):
        self._dtos.clear()


class FsHistoricalStore(HistoricalStore):
    """Store historical changes in a plain file from a operating system"""

    def __init__(self, filename: str):
        self._fn = filename
        self._has_dtos = None

    async def save_dtos(self, dtos: list[Metric]):
        async with aiofiles.open(self._fn, "a") as f:
            lines = [self._create_line(dto) for dto in dtos]
            await f.writelines(lines)

        self._has_dtos = True

    def _create_line(self, dto: Metric) -> str:
        quality = dto.properties["quality"].value if "quality" in dto.properties else ""
        return f"{dto.timestamp};{dto.value};{dto.datatype.name};{quality};{dto.alias};{dto.name}\n"

    async def load_all_dtos(self) -> list[Metric]:
        dtos = []
        async with aiofiles.open(self._fn, "r") as f:
            while True:
                line = await f.readline()

                values = line[:-1].split(";")

                if len(values) == 1:
                    break

                ts, value, datatype, quality, alias, name = values
                properties = (
                    PropertySet.from_kwargs(
                        quality=PropertyValue(int(quality), DataType.Int32)
                    )
                    if quality != ""
                    else PropertySet()
                )
                dtos.append(
                    Metric(
                        int(ts),
                        DataType[datatype].convert_value(value),
                        DataType[datatype],
                        int(alias),
                        name,
                        properties,
                        is_historical=True,
                    )
                )

        return dtos

    def has_dtos(self):
        if self._has_dtos is None:
            self._has_dtos = os.path.exists(self._fn)
        return self._has_dtos

    async def clear(self):
        os.remove(self._fn)
        self._has_dtos = False



from .. import Clock, UtcClock
from ..data import DataType, Metric
from ..mqtt import MqttClient, NodePayload, SpbMessage, Topic, Will


class _Seq:
    """Rotatory sequency number from 0 to 255"""

    def __init__(self):
        self._value = None

    def __call__(self) -> int:
        """If it is called, the previous value is increased and return it"""
        if self._value is None:
            self._value = 0
            return self._value

        self._value = 0 if self._value == 255 else self._value + 1
        return self._value

    @property
    def value(self) -> int:
        """Return their value (not increasing)"""
        if self._value is None:
            raise RuntimeError("Sequence not initialized")

        return self._value


class MqttNodeCarrier:
    """Helper to send Node messages to mqtt"""

    def __init__(
        self,
        mqtt_client: MqttClient,
        node_name: str,
        clock: Clock | None = None,
    ):
        self._mqtt_client = mqtt_client
        self._node_name = node_name
        self._clock = clock or UtcClock()
        self._seq = None
        self._bd_seq = _Seq()

    async def connect(self):
        """Connect node to MQTT Server"""
        ts = self._clock.timestamp()
        will = Will(
            SpbMessage(
                Topic.from_component(self._node_name, "NDEATH"),
                NodePayload(
                    timestamp=ts,
                    metrics=[Metric(ts, self._bd_seq(), DataType.Int64, name="bdSeq")],
                ),
            ),
            qos=1,
            retain=False,
        )
        self._seq = _Seq()

        await self._mqtt_client.connect(self._node_name, will=will)
        await self._mqtt_client.subscribe("spBv1.0/STATE/+", qos=1)
        await self._mqtt_client.subscribe(
            Topic.from_component(self._node_name, "NCMD").value, qos=1
        )

    async def send_birth(self, metrics: list[Metric]):
        """Send birth certificate"""
        if self._seq is None:
            raise RuntimeError("The node is not connected yet")

        ts = self._clock.timestamp()
        await self._mqtt_client.publish(
            SpbMessage(
                Topic.from_component(self._node_name, "NBIRTH"),
                NodePayload(
                    ts,
                    metrics
                    + [Metric(ts, self._bd_seq.value, DataType.Int64, name="bdSeq")],
                    self._seq(),
                ),
            ),
            qos=0,
            retain=False,
        )

    async def send_data_changes(self, metrics: list[Metric]):
        """Send metrics changes to MQTT Server"""
        if self._seq is None:
            raise RuntimeError("The node is not connected yet")

        await self._mqtt_client.publish(
            SpbMessage(
                Topic.from_component(self._node_name, "NDATA"),
                NodePayload(self._clock.timestamp(), metrics, self._seq()),
            ),
            qos=0,
            retain=False,
        )

    async def send_death(self):
        """Send death certificate to MQTT Server"""
        if self._seq is None:
            raise RuntimeError("The node is not connected yet")

        ts = self._clock.timestamp()
        await self._mqtt_client.publish(
            SpbMessage(
                Topic.from_component(self._node_name, "NDEATH"),
                NodePayload(
                    ts,
                    [Metric(ts, self._bd_seq.value, DataType.Int64, name="bdSeq")],
                ),
            ),
            qos=1,
            retain=False,
        )

    async def confirm_commands(self, writings: list[Metric]):
        """Send confirmation of commands to the MQTT Server"""
        if self._seq is None:
            raise RuntimeError("The node is not connected yet")

        metrics = []
        for writing in writings:
            if writing.timestamp == 0:
                metrics.append(
                    Metric(
                        self._clock.timestamp(),
                        writing.value,
                        writing.datatype,
                        writing.alias,
                        writing.name,
                        writing.properties,
                    )
                )
                continue
            metrics.append(writing)

        await self._mqtt_client.publish(
            SpbMessage(
                Topic.from_component(self._node_name, "NDATA"),
                NodePayload(self._clock.timestamp(), metrics, self._seq()),
            ),
            qos=0,
            retain=False,
        )

    async def disconnect(self):
        """Disconnect from MQTT Server"""
        await self._mqtt_client.disconnect()

    async def deliver_message(self) -> SpbMessage:
        """Return message content from MQTT Server"""
        return await self._mqtt_client.deliver_message()

    @property
    def is_connected(self):
        return self._mqtt_client.is_connected



import asyncio
import logging
from typing import Callable, Coroutine, Sequence

from aiospb import Clock, UtcClock
from aiospb.data import DataType, Metric, WriteRequest
from aiospb.mqtt import (
    HostPayload,
    MqttClient,
    MqttError,
    NodePayload,
    SpbMessage,
    Topic,
    Will,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class _Observer:
    def __init__(self, callback: Callable, filter: str = ""):
        self.callback = callback
        self._filter = filter

    async def notify(self, message: SpbMessage):
        if not self._filter or self._filter in message.topic.component_name:
            await self.callback(message)


class _MessageSorter:
    def __init__(self, node_name: str, clock: Clock, reorder_time: float = 2.0):
        self._name = node_name
        self._next_seq = 0
        self._clock = clock
        self._bd_seq = 0
        self._reorder_time = int(reorder_time * 1000)  # ms
        self._messages = {}
        self._timeout = 0

    def register_message(self, message: SpbMessage):
        content = message.payload.to_dict()
        if message.is_a("NCMD"):
            return

        if message.is_a("NDEATH"):
            self._messages[None] = message
            return

        seq = content["seq"]
        if message.is_a("NBIRTH"):
            self._messages.clear()
            self._next_seq = seq
            self._bd_seq = self._get_bd_seq(content["metrics"])

        self._messages[seq] = message

    def _get_bd_seq(self, metrics):
        for metric in metrics:
            if metric.get("name") == "bdSeq":
                return metric["value"]

    def nexts(self) -> list[SpbMessage]:
        output = []
        while self._messages:
            if self._next_seq in self._messages:
                output.append(self._messages.pop(self._next_seq))
                self._next_seq = 0 if self._next_seq == 255 else self._next_seq + 1
                self._timeout = self._clock.timestamp() + self._reorder_time
            elif None in self._messages:
                message = self._messages.pop(None)
                if message.payload.to_dict()["metrics"][0]["value"] != self._bd_seq:
                    continue

                output.append(message)
                self._messages.clear()
                self._next_seq = 0
            else:
                if self._clock.timestamp() > self._timeout:
                    self._messages.clear()
                    raise TimeoutError(
                        f"There are {len(self._messages)} messages disordered too much time"
                    )
                break
        return output


class _NodeCommander:
    def __init__(self, node_name, mqtt_client, clock):
        self._datatypes = {}
        self._aliases = {}
        self._clock = clock
        self._node_name = node_name
        self._mqtt_client = mqtt_client
        self._pending_commands = {}

    async def notify(self, message: SpbMessage):
        if message.is_a("NBIRTH"):
            metrics = message.payload.to_dict()["metrics"]

            self._datatypes.clear()
            for metric in metrics:
                datatype = (
                    DataType[metric["dataType"]]
                    if type(metric["dataType"]) is str
                    else DataType(metric["dataType"])
                )
                if "alias" in metric:
                    self._datatypes[metric["alias"]] = datatype
                    self._aliases[metric["name"]] = metric["alias"]
                self._datatypes[metric["name"]] = datatype
        elif message.is_a("NDATA"):
            metrics = message.payload.to_dict()["metrics"]
            self._clear_expired_commands()

            changes = [
                (data["value"], data.get("name", ""), data.get("alias", 0))
                for data in metrics
            ]

            for expected_changes, samples in self._pending_commands.values():
                if expected_changes == changes:
                    samples.set_result([Metric.from_dict(data) for data in metrics])

    def _clear_expired_commands(self):
        now = self._clock.timestamp()

        for ts in list(self._pending_commands.keys()):
            if now > ts:
                self._pending_commands.pop(ts)

    async def write_metrics(
        self, requests: Sequence[WriteRequest], timeout
    ) -> list[Metric]:
        logger.info(
            f'Requesting {len(requests)} metric changes to edge node "{self._node_name}"...'
        )
        changes = await self._send_command(requests)
        logger.debug("Command sent to edge node")
        if not changes:
            return []

        ts = self._clock.timestamp() + int(1000 * timeout)
        samples = asyncio.Future()
        self._pending_commands[ts] = (changes, samples)
        logger.debug(f"Waiting confirmation by node {self._node_name}...")
        metrics = await asyncio.wait_for(samples, timeout=timeout)
        logger.info(f'Recieved feedback of writing in "{self._node_name}"')
        for metric in metrics:
            logger.debug(f"Metric confirmation by {self._node_name} is: {metric}")
        return metrics

    async def _send_command(self, requests):
        ts = self._clock.timestamp()
        metrics = []
        for request in requests:
            alias = request.alias
            name = request.metric_name
            if alias not in self._aliases.values() and name not in self._datatypes:
                continue
            datatype = self._datatypes.get(alias) or self._datatypes[name]

            if name in self._aliases:
                alias = self._aliases[name]
                name = ""

            metrics.append(Metric(ts, request.value, datatype, alias, name))

        if not metrics:
            return []

        await self._mqtt_client.publish(
            SpbMessage(
                Topic.from_component(self._node_name, "NCMD"), NodePayload(ts, metrics)
            ),
            qos=0,
            retain=False,
        )
        return [(metric.value, metric.name, metric.alias) for metric in metrics]


class HostBridge:
    """Easy interface to host applications"""

    def __init__(
        self,
        hostname: str,
        mqtt_client: MqttClient,
        groups: str = "",
        reorder_timeout: float = 2,
        max_delay: int = 10,
        clock: Clock | None = None,
    ):
        self._hostname = hostname
        self._groups = groups.split(",")
        self._mqtt_client = mqtt_client
        self._reorder_timeout = reorder_timeout
        self._max_delay = max_delay
        self._clock = clock if clock else UtcClock()

        self._observers = []
        self._state = "offline"
        self._recieving = None
        self._listen_nodes = None
        self._sorters = {}
        self._commanders = {}

    @property
    def hostname(self) -> str:
        """Name of the host application"""
        return self._hostname

    @property
    def state(self):
        return self._state

    async def establish_session(self, groups: list[str] | None = None):
        """Init session to listen edge nodes"""

        await self._mqtt_client.connect(
            self._hostname,
            will=Will(
                SpbMessage(
                    Topic.from_component(self._hostname, "STATE"),
                    HostPayload(self._clock.timestamp(), False),
                ),
                qos=1,
                retain=True,
            ),
        )
        groups = groups or ["+"]
        for group in groups:
            await self._mqtt_client.subscribe(f"spBv1.0/{group}/+/+", qos=1)

        self._recieving = asyncio.create_task(self._recieve_node_messages())

        await self._mqtt_client.publish(
            SpbMessage(
                Topic.from_component(self._hostname, "STATE"),
                HostPayload(self._clock.timestamp(), True),
            ),
            qos=1,
            retain=True,
        )
        logger.info(f'Host application "{self._hostname}" has established session')
        self._state = "online"

    async def terminate_session(self):
        """Close cleanly a session"""

        if self._recieving is not None:
            self._recieving.cancel()

        try:
            await self._mqtt_client.publish(
                SpbMessage(
                    Topic.from_component(self._hostname, "STATE"),
                    HostPayload(self._clock.timestamp(), False),
                ),
                qos=1,
                retain=True,
            )
            await self._mqtt_client.disconnect()
        except MqttError as e:
            logger.error("MQTT connection broken when terminating session")
            logger.exception(e)

    def done(self):
        return self._listen_nodes is None or self._listen_nodes.done()

    def observe_nodes(
        self,
        callback: Callable[[SpbMessage], Coroutine[None, None, None]],
        node_filter: str = "",
    ) -> None:
        """Add one callable observer when it rece"""

        self._observers.append(_Observer(callback, node_filter))

    async def _recieve_node_messages(self):
        while True:
            try:
                message: SpbMessage = await self._mqtt_client.deliver_message()
                if not message.is_from_node():
                    continue

                node_name = message.topic.component_name
                if node_name not in self._sorters:
                    self._commanders[node_name] = commander = _NodeCommander(
                        node_name, self._mqtt_client, self._clock
                    )
                    self._observers.append(_Observer(commander.notify, node_name))

                    self._sorters[node_name] = _MessageSorter(
                        node_name, self._clock, self._reorder_timeout
                    )
                sorter = self._sorters[node_name]
                sorter.register_message(message)

                try:
                    for message in sorter.nexts():
                        results = await asyncio.gather(
                            *[observer.notify(message) for observer in self._observers],
                            return_exceptions=True,
                        )
                        for result, observer in zip(results, self._observers):
                            if type(result) is Exception:
                                logger.warning(
                                    f"Observer {observer} raised exception when notified"
                                )
                                logger.exception(result)
                except TimeoutError:
                    ts = self._clock.timestamp()
                    await self._mqtt_client.publish(
                        SpbMessage(
                            Topic.from_component(node_name, "NCMD"),
                            NodePayload(
                                ts,
                                metrics=[
                                    Metric(
                                        ts,
                                        True,
                                        DataType.Boolean,
                                        name="Node Control/Rebirth",
                                    )
                                ],
                            ),
                        ),
                        qos=0,
                        retain=False,
                    )

            except Exception as e:
                logger.error("Task of recieving node messages has stopped")
                logger.exception(e)
                raise e

    async def write_metrics(
        self, node_name: str, write_requests: list[WriteRequest], timeout: float = 10.0
    ):
        """Request changes to metrics"""

        try:
            return await self._commanders[node_name].write_metrics(
                write_requests, timeout
            )

        except TimeoutError as e:
            logger.warning(
                f"Not recieved confirmation of writing for {timeout} seconds"
            )
            raise e
        except Exception as e:
            logger.error(
                f"Request of metrics changes at {node_name} has been stopped by exception"
            )
            logger.exception(e)
            raise e



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Self, TypeAlias


class Quality(Enum):
    BAD = 0
    GOOD = 192
    STALE = 500


ValueType: TypeAlias = bool | int | float | str | bytes | None


class DataType(Enum):
    Unknown = 0
    Int8 = 1
    Int16 = 2
    Int32 = 3
    Int64 = 4
    UInt8 = 5
    UInt16 = 6
    UInt32 = 7
    UInt64 = 8
    Float = 9
    Double = 10
    Boolean = 11
    String = 12
    DateTime = 13
    Text = 14
    UUID = 15
    DataSet = 16
    Bytes = 17
    File = 18
    Template = 19
    PropertySet = 20
    PropertySetList = 21
    Int8Array = 22
    Int16Array = 23
    Int32Array = 24
    Int64Array = 2
    UInt8Array = 26
    UInt16Array = 27
    UInt32Array = 28
    UInt64Array = 29
    FloatArray = 30
    DoubleArray = 31
    BooleanArray = 32
    StringArray = 33
    DateTimeArray = 34

    @classmethod
    def for_(cls, value: ValueType) -> Self:
        v_type = type(value)
        if v_type not in _DEFAULT_DTs:
            raise ValueError(f"There is no default for type {v_type}")
        return _DEFAULT_DTs[v_type]

    def convert_value(self, str_value: str) -> ValueType:
        if str_value == "None":
            return

        if "Int" in self.name:
            return int(str_value)

        if "Float" in self.name or "Double" in self.name:
            return float(str_value)

        if "Boolean" == self.name:
            return str_value == True

        if "Bytes" == self.name:
            return str_value.encode()

        return str_value


_DEFAULT_DTs = {
    int: DataType.Int64,
    float: DataType.Float,
    bool: DataType.Boolean,
    str: DataType.String,
    bytes: DataType.Bytes,
}


@dataclass(frozen=True)
class PropertyValue:
    value: ValueType
    data_type: DataType

    def to_dict(self) -> dict[str, ValueType]:
        return {"value": self.value, "dataType": self.data_type.name}

    @classmethod
    def from_dict(cls, dump: dict[str, Any]) -> Self:
        datatype = (
            DataType[dump["dataType"]]
            if type(dump["dataType"]) is str
            else DataType(dump["dataType"])
        )
        return cls(dump["value"], datatype)

    @classmethod
    def from_value(cls, value: ValueType) -> Self:
        value_type = type(value)
        if value_type not in cls._DEFAULTS:
            raise ValueError(f"Not default for type {type(value)}")

        datatype = cls._DEFAULTS[value_type]

        return cls(value, datatype)


@dataclass(frozen=True)
class PropertySet:
    """Map or properties"""

    keys: tuple[str, ...] = field(default_factory=tuple)
    values: tuple[PropertyValue, ...] = field(default_factory=tuple)

    def __contains__(self, key: str) -> bool:
        return key in self.keys

    def __getitem__(self, key: str) -> PropertyValue:
        try:
            index = self.keys.index(key)
        except ValueError:
            raise KeyError(f"Key {key} not found in properties")

        return self.values[index]

    def __bool__(self) -> bool:
        return bool(self.keys)

    @classmethod
    def from_dict(cls, dump: dict[str, Any]) -> Self:
        """Construct a property set from a dict"""
        keys = tuple(sorted(dump.keys()))
        return cls(keys, tuple([PropertyValue.from_dict(dump[key]) for key in keys]))

    @classmethod
    def from_kwargs(cls, **kwargs) -> Self:
        """Construct property set from keywords arguments"""
        keys = tuple(sorted(kwargs.keys()))

        return cls(keys, tuple([kwargs[key] for key in keys]))

    def as_dict(self) -> dict[str, dict[str, Any]]:
        """Convert object to a dict"""
        return {key: value.to_dict() for key, value in zip(self.keys, self.values)}


@dataclass
class Metric:
    """Data Transfer Object (dto) to be included in Sparkplug Messages"""

    timestamp: int
    value: ValueType
    datatype: DataType
    alias: int = 0
    name: str = ""
    properties: PropertySet = field(default_factory=PropertySet)
    is_transient: bool = False
    is_historical: bool = False

    @classmethod
    def from_dict(cls, dump: dict[str, Any]) -> Self:
        """Construct dto from a dict"""
        return cls(
            dump["timestamp"],
            dump["value"],
            DataType[dump["dataType"]],
            dump.get("alias", 0),
            dump.get("name", ""),
            PropertySet.from_dict(dump.get("properties", {})),
            dump.get("is_transient", False),
            dump.get("is_historical", False),
        )

    def as_dict(self) -> dict[str, Any]:
        """Convert dto to dict for parsing"""
        dump = {
            "timestamp": self.timestamp,
            "value": self.value,
            "dataType": self.datatype.name,
        }
        if self.alias:
            dump["alias"] = self.alias

        if self.name:
            dump["name"] = self.name

        if self.is_transient:
            dump["is_transient"] = True

        if self.properties:
            dump["properties"] = self.properties.as_dict()

        if self.is_historical:
            dump["is_historical"] = True
        return dump


@dataclass(frozen=True)
class WriteRequest:
    value: ValueType
    metric_name: str = ""
    alias: int = 0
